#include "cuda_fx.h"

/*
 * Computes CCI on the CPU
 * Parameters:
 *  float* highValues. host array of high values.
 *  float* lowValues. host array of low values.
 *  float* closeValues. host array of close values.
 *  float* spreadValues. host array of spread values.
 *  int nRecords. The length of the above arrays.
 *  int takePositionAfterNRecords. The index after which we're going to take the first position.
 *  int minN & int maxN. The lowest and highest parameter values for the SMATP that are to be evaluated.
 *  int minUpperBreakout and int maxUpperBreakout. The range over which the upper breakout point will be evaluated.
 *  int minLowerBreakout and int maxLowerBreakout. The range over which the lower breakout point will be evaluated.
 *  int& chosenN and float& profit. The result: the chosen N and the profit for that N.
 *
 * Task:
 *  Tries all values from <minN> to <maxN> for the SMATP, <minUpperBreakout> to <maxUpperBreakout> for the upper (BUY)
 *  breakout point and <minLowerBreakout> to <maxLowerBreakout> for the lower (SELL) breakout point and determines which is best.
 *  Results are written to <chosenN> and <profit>.
 */
void cpuCciMain(float* highValues, float* lowValues, float* closeValues, float* spreadValues, int nRecords, int takePositionAfterNRecords, int minN, int maxN, int minUpperBreakout, int maxUpperBreakout, int minLowerBreakout, int maxLowerBreakout, int& chosenN, int& chosenUpperBreakout, int& chosenLowerBreakout, float& profit) {
	int intMAPointCount, intUpperBreakout, intLowerBreakout, intSampleIndex, i, intLotPosition, intOldLotPosition;
	float fltProfit, fltTPMA, fltMD;
	float *fltTP = (float *) malloc (sizeof(float) * nRecords);
	float *fltCci = (float *) malloc (sizeof(float) * nRecords);
	profit = -pow(2.0f, (float) (sizeof(float) * 8) - 1);

	// Pre-compute the TP values
	for (intSampleIndex = 0; intSampleIndex < nRecords; intSampleIndex++)
		fltTP[intSampleIndex] = (highValues[intSampleIndex] + lowValues[intSampleIndex] + closeValues[intSampleIndex]) / 3.0f;


	// Loop over values for number of samples to use in the MA
	for (intMAPointCount = minN; intMAPointCount<= maxN; intMAPointCount++) {
		// Initialize variables
		fltTPMA = 0;

		// Initialize MA using samples that occur prior to takePositionAfterNRecords
		for (intSampleIndex = takePositionAfterNRecords - intMAPointCount - 1; intSampleIndex < takePositionAfterNRecords - 1; intSampleIndex++)
			fltTPMA += fltTP[intSampleIndex];
		
		fltTPMA /= intMAPointCount;

		// Loop over remaining samples, updating TPMA and MD in order to pre-compute CCI
		for (; intSampleIndex < nRecords; intSampleIndex++) {
			fltTPMA += (fltTP[intSampleIndex] - fltTP[intSampleIndex - intMAPointCount]) / (float) intMAPointCount;
			
			// Calculate the Mean Deviation based on the new TPMA value
			fltMD = 0;
			for (i = intSampleIndex - intMAPointCount + 1; i <= intSampleIndex; i++)
				fltMD += abs(fltTP[i] - fltTPMA);

			fltMD /= (float) intMAPointCount;

			fltCci[intSampleIndex] = (fltTP[intSampleIndex] - fltTPMA) / (0.015f * fltMD);
		}

		// Now, vary the upper and lower breakout thresholds
		for (intUpperBreakout = minUpperBreakout; intUpperBreakout <= maxUpperBreakout; intUpperBreakout++) {
			for (intLowerBreakout = minLowerBreakout; intLowerBreakout >= maxLowerBreakout; intLowerBreakout--) {

				// Loop over the pre-computed CCI values assessing the threshold crossovers
				fltProfit = 0;
				intOldLotPosition = 0;
				intLotPosition = 0;
				for (intSampleIndex = takePositionAfterNRecords - 1; intSampleIndex < nRecords - 1; intSampleIndex++) {
					intLotPosition = fltCci[intSampleIndex] - intUpperBreakout > 0 ? 1 : (fltCci[intSampleIndex] - intLowerBreakout > 0 ? 0 : -1);
					fltProfit += (intOldLotPosition - intLotPosition) * (closeValues[intSampleIndex] + (intOldLotPosition - intLotPosition < 0 ? 1 : -1) *  0.5f * spreadValues[intSampleIndex]);
					intOldLotPosition = intLotPosition;

				}

				// Close final position	
				fltProfit += intLotPosition * closeValues[nRecords-1] - (intLotPosition==0?0: 0.5f * spreadValues[nRecords-1]);

				// If this value is the highest so far, save it
				if (fltProfit > profit) {
					profit = fltProfit;
					chosenN = intMAPointCount;
					chosenUpperBreakout = intUpperBreakout;
					chosenLowerBreakout = intLowerBreakout;
				}
			}
		}
	}

	// Free memory
	free(fltTP);
	free(fltCci);
}
