#include "cuda_fx.h"

/*
 * Computes CCI on the CPU. This is the optimized implementation.
 * Parameters:
 *  float* highValues. host array of high values.
 *  float* lowValues. host array of low values.
 *  float* closeValues. host array of close values.
 *  float* spreadValues. host array of spread values.
 *  int nRecords. The length of the above arrays.
 *  int takePositionAfterNRecords. The index after which we're going to take the first position.
 *  int minN & int maxN. The lowest and highest parameter values for the SMATP that are to be evaluated.
 *  int minUpperBreakout and int maxUpperBreakout. The range over which the upper breakout point will be evaluated.
 *  int minLowerBreakout and int maxLowerBreakout. The range over which the lower breakout point will be evaluated.
 *  int& chosenN and float& profit. The result: the chosen N and the profit for that N.
 *
 * Task:
 *  Tries all values from <minN> to <maxN> for the SMATP, <minUpperBreakout> to <maxUpperBreakout> for the upper (BUY)
 *  breakout point and <minLowerBreakout> to <maxLowerBreakout> for the lower (SELL) breakout point and determines which is best.
 *  Results are written to <chosenN> and <profit>.
 */
void cpuCciMain2(float* highValues, float* lowValues, float* closeValues, float* spreadValues, int nRecords, int takePositionAfterNRecords, int minN, int maxN, int minUpperBreakout, int maxUpperBreakout, int minLowerBreakout, int maxLowerBreakout, int& chosenN, int& chosenUpperBreakout, int& chosenLowerBreakout, float& profit) {
  float* cci = (float*) malloc(nRecords * sizeof(float)); // cache for cci
  float* tp = (float*) malloc(nRecords * sizeof(float)); // cache for Typical Price

  // compute TP values
  M_repI(nRecords) tp[i] = (highValues[i] + lowValues[i] + closeValues[i])/3;

  // outer loop, over all possibilities of N for SMATP. With argmax.
  chosenN = -1000000; profit = -1000000;
  for (int smatpN=minN; smatpN<=maxN; smatpN++){

    // calculate cci
    M_repI2(takePositionAfterNRecords-1, nRecords){
      // calc smatp
      float temp=0;
      for (int j=i-smatpN+1; j<=i; j++) temp += tp[j];
      float smatp = temp/smatpN;

      // calc md
      temp=0;
      for (int j=i-smatpN+1; j<=i; j++) temp += abs(tp[j]-smatp);
      float md = temp/smatpN;

      cci[i] = (tp[i]-smatp) / (float) (.015*md);
    }

    // find optimal upperBreakout (UB) value
    int bestUB=-1000000; float profit_bestUB = -1000000;
    for (int UB=minUpperBreakout; UB<=maxUpperBreakout; UB++){
      bool havePosition = cci[takePositionAfterNRecords-1] > UB;
      float cash = havePosition ? -closeValues[takePositionAfterNRecords-1] - spreadValues[takePositionAfterNRecords-1]/2 : 0;
      // simulation loop
      M_repI2(takePositionAfterNRecords, nRecords-1){
	if ((cci[i]>UB) ^ havePosition){
	  // then change position
	  cash += (havePosition ? closeValues[i] : -closeValues[i]) - spreadValues[i]/2;
	  havePosition = !havePosition;
	}
      }
      if (havePosition) cash += closeValues[nRecords-1] - spreadValues[nRecords-1]/2; // close final position
      if (cash>profit_bestUB) {profit_bestUB = cash; bestUB = UB;}
    }

    // find optimal lowerBreakout (LB) value
    int bestLB = -1000000; float profit_bestLB = -1000000;
    int LBFrom = min(minLowerBreakout, maxLowerBreakout), LBTo = max(minLowerBreakout, maxLowerBreakout); // just to have this unambiguous
    for (int LB = LBFrom; LB<=LBTo; LB++){
      bool havePosition = cci[takePositionAfterNRecords-1] < LB;
      float cash = havePosition ? closeValues[takePositionAfterNRecords-1] - spreadValues[takePositionAfterNRecords-1]/2 : 0;
      // simulation loop
      M_repI2(takePositionAfterNRecords, nRecords-1)
	if ((cci[i]<LB) ^ havePosition){
	  // then change position
	  cash += (havePosition ? -closeValues[i] : closeValues[i]) - spreadValues[i]/2;
	  havePosition = !havePosition;
	}
      if (havePosition) cash += -closeValues[nRecords-1] -spreadValues[nRecords-1]/2; // close final position
      if (cash>profit_bestLB) {profit_bestLB = cash; bestLB = LB;}
    }

    if (profit_bestLB + profit_bestUB > profit) {
      chosenN = smatpN;
      profit = profit_bestLB + profit_bestUB;
      chosenLowerBreakout = bestLB;
      chosenUpperBreakout = bestUB;
    }
      
  }
  free(cci); free(tp);
}

