#include "cuda_fx.h"

/*
 * Computes EMA-S and EMA-OC on the CPU
 * Parameters:
 *  bool checkOC. True if running EMA-OC, false if running EMA-S.
 *  float* openValues. host array of open values.
 *  float* closeValues. host array of close values.
 *  float* spreadValues. host array of spread values.
 *  int nRecords. The length of the above arrays.
 *  int takePositionAfterNRecords. The index after which we're going to take the first position.
 *  int minN & int maxN. The lowest and highest parameter values that are to be evaluated.
 *  int& chosenN and float& profit. The result: the chosen N and the profit for that N.
 *
 * Task:
 *  Tries all values from <minN> up to and including <maxN> and determines which is best.
 *  Results are written to <chosenN> and <profit>.
 */
void cpuEmaMain(bool checkOC, float* openValues, float* closeValues, float* spreadValues, int nRecords, int takePositionAfterNRecords, int minN, int maxN, int& chosenN, float& profit) {  
	int intMAPointCount, intSampleIndex, intLotPosition;
	float fltProfit, fltEMA, alpha;
	profit = -pow(2.0f, (float) (sizeof(float) * 8) - 1);

	// Loop from min analysis samples to max analysis samples
	for (intMAPointCount = minN; intMAPointCount<= maxN; intMAPointCount++) {
		// Initialize moving average & alpha
		fltEMA = closeValues[0];
		alpha = 2.0f / (float) (intMAPointCount + 1);

		// Initialize EMA using samples that occur prior to takePositionAfterNRecords
		for (intSampleIndex = 1; intSampleIndex < takePositionAfterNRecords - 1; intSampleIndex++)
			fltEMA += alpha * (closeValues[intSampleIndex] - fltEMA);
		
		// Take initial position & initial profit
		intLotPosition = (closeValues[intSampleIndex] > fltEMA ? 1 : -1);
		fltProfit = -intLotPosition * closeValues[intSampleIndex] - 0.5f * spreadValues[intSampleIndex];

		// Loop over remaining samples, updating EMA and changing position when necessary
		for (; intSampleIndex < nRecords - 1; intSampleIndex++) {
			fltEMA += alpha * (closeValues[intSampleIndex] - fltEMA);

			// Determine if current position still valid
			if ((closeValues[intSampleIndex] - fltEMA) * intLotPosition < 0 && (!checkOC || (openValues[intSampleIndex] - fltEMA) * intLotPosition < 0)) {
				// If not, change the position
				intLotPosition = -intLotPosition;
				fltProfit += -intLotPosition * 2 * closeValues[intSampleIndex] - spreadValues[intSampleIndex];
			}
		}

		// Close final position
		fltProfit += intLotPosition * closeValues[nRecords-1] - 0.5f * spreadValues[nRecords-1];

		// If this value is the highest so far, save it
		if (fltProfit > profit) {
			profit = fltProfit;
			chosenN = intMAPointCount;
		}
	}
}

/*
 * Computes EMA-D on the CPU
 * Parameters:
 *  float* closeValues. host array of close values.
 *  float* spreadValues. host array of spread values.
 *  int nRecords. The length of the above arrays.
 *  int takePositionAfterNRecords. The index after which we're going to take the first position.
 *  int minN & int maxN. The lowest and highest parameter values that are to be evaluated.
 *  int& chosenLongN, int& chosenShortN and float& profit. The result: the chosen N for the long curve,
 *		short curve and the profit for these N values.
 *
 * Task:
 *  Tries all values from <minN + 1> up to and including <maxN> for the long curve and <minN> up to the long curve
 *    N to determine which are best.
 *  Results are written to <chosenLongN>, <chosenShortN> and <profit>.
 */
void cpuDoubleEmaMain(float* closeValues, float* spreadValues, int nRecords, int takePositionAfterNRecords, int minN, int maxN, int& chosenLongN, int& chosenShortN, float& profit) {  
	int intMAPointCountLong, intMAPointCountShort, intSampleIndex, intLotPosition;
	float fltProfit, fltEMALong, fltEMAShort, alphaLong, alphaShort;
	profit = -pow(2.0f, (float) (sizeof(float) * 8) - 1);

	// Loop from min analysis samples + 1 to max analysis samples - long curve
	for (intMAPointCountLong = minN+1; intMAPointCountLong<= maxN; intMAPointCountLong++) {
		// Initialize moving average & alpha
		alphaLong = 2.0f / (intMAPointCountLong + 1);

		// Loop from min analysis samples to long curve sample count - short curve
		for (intMAPointCountShort = minN; intMAPointCountShort < intMAPointCountLong; intMAPointCountShort++) {
			// Initialize moving average & alpha
			alphaShort = 2.0f / (intMAPointCountShort + 1);
			fltEMALong = closeValues[0];
			fltEMAShort = closeValues[0];
			
			// Initialize EMA using samples that occur prior to takePositionAfterNRecords
			for (intSampleIndex = 1; intSampleIndex < takePositionAfterNRecords - 1; intSampleIndex++) {
				fltEMALong += alphaLong * (closeValues[intSampleIndex] - fltEMALong);
				fltEMAShort += alphaShort * (closeValues[intSampleIndex] - fltEMAShort);
			}
			
			// Take initial position & initial profit
			intLotPosition = (fltEMAShort >= fltEMALong ? 1 : -1);
			fltProfit = -intLotPosition * closeValues[intSampleIndex] - 0.5f * spreadValues[intSampleIndex];

			// Loop over remaining samples, updating EMA and changing position when necessary
			for (; intSampleIndex < nRecords - 1; intSampleIndex++) {
				fltEMALong += alphaLong * (closeValues[intSampleIndex] - fltEMALong);
				fltEMAShort += alphaShort * (closeValues[intSampleIndex] - fltEMAShort);

				// Determine if current position still valid
				if ((fltEMAShort - fltEMALong) * intLotPosition < 0) {
					// If not, change the position
					intLotPosition = -intLotPosition;
					fltProfit += -intLotPosition * 2 * closeValues[intSampleIndex] - spreadValues[intSampleIndex];
				}
			}

			// Close final position
			fltProfit += intLotPosition * closeValues[nRecords-1] - 0.5f * spreadValues[nRecords-1];

			// If this value is the highest so far, save it
			if (fltProfit > profit) {
				profit = fltProfit;
				chosenLongN = intMAPointCountLong;
				chosenShortN = intMAPointCountShort;
			}
		}
	}
}