#include "cuda_fx.h"


/*
 * Computes MACD on the CPU
 * Parameters:
 *  float* closeValues. host array of close values.
 *  float* spreadValues. host array of spread values.
 *  int nRecords. The length of the above arrays.
 *  int takePositionAfterNRecords. The index after which we're going to take the first position.
 *  int minN & int maxN. The lowest and highest parameter values that are to be evaluated.
 *  int& chosenLongN, int& chosenShortN, int& chosenSignalN and float& profit. The result: the chosen N for the 
 *      long curve, short curve, signal curve and the profit for these N values.
 *
 * Task:
 *  Tries all values from <minN + 1> up to and including <maxN> for the long curve and <minN> up to the long curve
 *    N as well as from <minN> to <maxN> for the signal curve to determine which are best.
 *  Results are written to <chosenLongN>, <chosenShortN>, <chosenSignalN> and <profit>.
 */
void cpuMacdMain(float* closeValues, float* spreadValues, int nRecords, int takePositionAfterNRecords, int minN, int maxN, int& chosenLongN, int& chosenShortN, int& chosenSignalN, float& profit) {  
	int intMAPointCountLong, intMAPointCountShort, intMAPointCountSignal, intSampleIndex, intLotPosition;
	float fltProfit, fltEMALong, fltEMAShort, fltEMASignal, alphaLong, alphaShort, alphaSignal;
	float *MACD = (float *) malloc (sizeof(float) * nRecords);
	profit = -pow(2.0f, (float) (sizeof(float) * 8) - 1);

	// Loop from min analysis samples + 1 to max analysis samples - long curve
	for (intMAPointCountLong = minN+1; intMAPointCountLong<= maxN; intMAPointCountLong++) {
		// Initialize moving average & alpha
		alphaLong = 2.0f / (intMAPointCountLong + 1);

		// Loop from min analysis samples to long curve sample count - short curve
		for (intMAPointCountShort = minN; intMAPointCountShort < intMAPointCountLong; intMAPointCountShort++) {
			// Initialize alpha & MACD values
			alphaShort = 2.0f / (intMAPointCountShort + 1);
			fltEMALong = closeValues[0];
			fltEMAShort = closeValues[0];
			MACD[0] = 0;
				
			// Pre-compute MACD before looking at signal line values
			for (intSampleIndex = 1; intSampleIndex < nRecords; intSampleIndex++) {
				fltEMALong += alphaLong * (closeValues[intSampleIndex] - fltEMALong);
				fltEMAShort += alphaShort * (closeValues[intSampleIndex] - fltEMAShort);
				MACD[intSampleIndex] = fltEMAShort - fltEMALong;
			}

			// Loop across entire range of possible MA values
			for (intMAPointCountSignal = minN; intMAPointCountSignal <= maxN; intMAPointCountSignal++) {
				// Initialize alpha & MAs
				alphaSignal = 2.0f / (intMAPointCountSignal + 1);
				fltEMASignal = 0;
			
				// Initialize EMA using samples that occur prior to takePositionAfterNRecords
				for (intSampleIndex = 1; intSampleIndex < takePositionAfterNRecords - 1; intSampleIndex++) {
					fltEMASignal += (float) (alphaSignal * (MACD[intSampleIndex] - fltEMASignal));
				}
				
				// Take initial position & initial profit
				intLotPosition = (MACD[intSampleIndex] >= fltEMASignal ? 1 : -1);
				fltProfit = -intLotPosition * closeValues[intSampleIndex] - 0.5f * spreadValues[intSampleIndex];

				// Loop over remaining samples, updating EMA and changing position when necessary
				for (; intSampleIndex < nRecords - 1; intSampleIndex++) {
					fltEMASignal += (float) (alphaSignal * (MACD[intSampleIndex] - fltEMASignal));

					// Determine if current position still valid
					if ((MACD[intSampleIndex] - fltEMASignal) * intLotPosition < 0) {
						// If not, change the position
						intLotPosition = -intLotPosition;
						fltProfit += -intLotPosition * 2 * closeValues[intSampleIndex] - spreadValues[intSampleIndex];
					}
				}

				// Close final position
				fltProfit += intLotPosition * closeValues[nRecords-1] - 0.5f * spreadValues[nRecords-1];

				// If this value is the highest so far, save it
  				if (fltProfit > profit) {
					profit = fltProfit;
					chosenLongN = intMAPointCountLong;
					chosenShortN = intMAPointCountShort;
					chosenSignalN = intMAPointCountSignal;
				}
			}
		}
	}
}