#include "cuda_fx.h"

/*
Notes & ideas.

to achieve maximum compute utilization, it's important to have at least as many blocks as there are multiprocessors, i.e. at least 30.

There are multiple possible bottlenecks:
 loading data to Smem; computing the profit following a given chart indicator's advice; and argmax at the end.
 I'm guessing the profit computation is going to be the bottleneck. Let's do the other parts as simple as possible, for now.

argmax: the first phase can be done simply as the last action by the computation threads.
 
for efficient loading of shared memory, the block size should be a multiple of 32.

*/


// ---------------------------------------------------------------------------------------------------- simple EMA: detect close price moving over EMA.


/*The EMA kernel. this computes the profit/loss for a particular parameter setting, and outputs that profit/loss into a Gmem array.
  parameters:
   bool checkOC. whether to do full OC check or not.
   float* globalOpen is a Gmem array of the open values.
   float* globalClose is the global memory array of close values.
   float* globalSpread is the global memory array of spread values.
   int nRecords is the number of close values (or, equivalently, the number of spread values).
   int* parameterValues are the parameter values for which we wish to run the simulation. Perhaps simply 2,3,...,N-1 or so.
   int nParameterValues indicates the length of parameterValues.
   float* out is a global memory array, of length nParameterValues, where the final profit/loss for every parameter setting is to be recorded.
*/

template <bool checkOC>
__global__ void gpuEmaKernel(float* globalOpen, float* globalClose, float* globalSpread, int nRecords, int takePositionAfterNRecords, int* parameterValues, int nParameterValues, float* out){
  // find out which thread we are
  int localTid = threadIdx.x;
  int bid = blockIdx.x;
  int globalTid = localTid + bid * blockDim.x;

  // declaring shared memory. I use "extern __shared__", and split it up in the right pieces.
  extern __shared__ float sharedMemData[];
  float* close = sharedMemData;
  float* spread = sharedMemData + nRecords;
  float* open = sharedMemData + nRecords*2; // ignored if !checkOC

  // loading the data into shared memory. these will be coalesced Gmem reads if block size is a multiple of 32.
  for (int i=localTid; i<nRecords; i+=blockDim.x){
    close[i] = globalClose[i];
    spread[i] = globalSpread[i];
    if (checkOC) open[i] = globalOpen[i];
  }
  __syncthreads(); // wait until all threads have done their bit of Smem loading.

  // now, after doing their part of the data loading, and crucially after the syncthreads, unneeded threads quit, and needed threads calc alpha.
  if (globalTid>=nParameterValues) return;
  float alpha = 2.0/(parameterValues[globalTid]+1);
  
  // initial segment of the sequence. no position is taken - we just compute the ema.
  float ema = close[0];
  M_repI2(1, takePositionAfterNRecords){
    ema += alpha * (close[i]-ema);
  }

  // set up initial position and cash counting.
  // I use floats because those are faster than ints, on GPUs (at least when multiplying with floats)
  float position = (close[takePositionAfterNRecords-1]>=ema) ? 1 : -1;
  float cashSoFar = - position*close[takePositionAfterNRecords-1] - spread[takePositionAfterNRecords-1]/2;
  
  // the main computation loop
  M_repI2(takePositionAfterNRecords, nRecords-1){
    float priceNow = close[i];
    ema += alpha * (priceNow - ema);
    if ((priceNow-ema) * position <0 && (!checkOC || (open[i]-ema)*position<0)){ // then change the position
      position = -position;
      cashSoFar -= position*2 * priceNow + spread[i];
    }
  }

  // close final position
  cashSoFar += position * close[nRecords-1] - spread[nRecords-1]/2;

  // write result
  out[globalTid] = cashSoFar;
}


/*
  Parameters:
   bool checkOC. if true: do EMA-OC. if false: do EMA-S.
   float* d_openValues. device array of open values.
   float* d_closeValues. device array of close values.
   float* d_spreadValues. device array of spread values.
   int nRecords. The length of the above arrays.
   int takePositionAfterNRecords. The index after which we're going to take the first position.
   int minN & int maxN. The lowest and highest parameter values that are to be evaluated.
   int& chosenN and float& profit. The result: the chosen N and the profit for that N.

  Task:
   Tries all values from <minN> up to and including <maxN> and determines which is best.
   Results are written to <chosenN> and <profit>.
*/   
void gpuEmaMain(bool checkOC, float* d_openValues, float* d_closeValues, float* d_spreadValues, int nRecords, int takePositionAfterNRecords, int minN, int maxN, int& chosenN, float& profit){  
  // prep parameter values array in device memory
  int nParameterValues = maxN-minN+1;
  int* h_parameterValues = (int*) malloc(nParameterValues*sizeof(int));
  for (int i=0; i<nParameterValues; i++) h_parameterValues[i] = i+minN;
  M_cudaMalloc2(d_parameterValues, int, nParameterValues);
  M_cudaCall(cudaMemcpy(d_parameterValues, h_parameterValues, nParameterValues*sizeof(int), cudaMemcpyHostToDevice));
  
  // kernel execution
  int blockSize = 32; // i don't know what's best, but 32 is my first guess.
  int nBlocks = (nParameterValues+blockSize-1) / blockSize;
  int sharedMemSize = nRecords * (checkOC?3:2) * sizeof(float);
  M_cudaMalloc2(d_profit, float, nParameterValues);
  if (checkOC) {M_cudaCall((gpuEmaKernel<true><<<nBlocks, blockSize, sharedMemSize>>>(d_openValues, d_closeValues, d_spreadValues, nRecords, takePositionAfterNRecords, d_parameterValues, nParameterValues, d_profit)));}
  else M_cudaCall((gpuEmaKernel<false><<<nBlocks, blockSize, sharedMemSize>>>(d_openValues, d_closeValues, d_spreadValues, nRecords, takePositionAfterNRecords, d_parameterValues, nParameterValues, d_profit)));

  // argmax on host (i'll later replace this with argmax on device).
  float* h_profit = (float*) malloc(nParameterValues * sizeof(float));
  M_cudaCall(cudaMemcpy(h_profit, d_profit, nParameterValues*sizeof(float), cudaMemcpyDeviceToHost));
  int chosenSettingId = cpuArgmax(h_profit, nParameterValues);
  chosenN = h_parameterValues[chosenSettingId];
  profit = h_profit[chosenSettingId];
  

  // debugging output
  //M_repI(nParameterValues) printf("N=%d --> profit = %f.\n", h_parameterValues[i], h_profit[i]);
  
  // finalization
  free(h_parameterValues);
  cudaFree(d_parameterValues);
  free(h_profit);
  cudaFree(d_profit);
}




// ---------------------------------------------------------------------------------------------------- double EMA: detect short EMA crossing over long EMA
// this is copied & modified from the simple EMA

/*The EMA kernel. this computes the profit/loss for a particular parameters setting, and outputs that profit/loss into a Gmem array.
  parameters:
   float* globalClose is the global memory array of close values.
   float* globalSpread is the global memory array of spread values.
   int nRecords is the number of close values (or, equivalently, the number of spread values).
   int* longN, and int* shortN, are the parameter values for which we wish to run the simulation. 
   int nParameterValues indicates the length of longN and shortN.
   float* out is a global memory array, of length nParameterValues, where the final profit/loss for every parameter setting is to be recorded.
*/
__global__ void gpuDoubleEmaKernel(float* globalClose, float* globalSpread, int nRecords, int takePositionAfterNRecords, int* longN, int* shortN, int nParameterValues, float* out){
  // find out which thread we are
  int localTid = threadIdx.x;
  int bid = blockIdx.x;
  int globalTid = localTid + bid * blockDim.x;

  // declaring shared memory. I use "extern __shared__", and split it up in the right pieces.
  extern __shared__ float sharedMemData[];
  float* close = sharedMemData;
  float* spread = sharedMemData + nRecords;

  // loading the data into shared memory. these will be coalesced Gmem reads if block size is a multiple of 32.
  for (int i=localTid; i<nRecords; i+=blockDim.x){
    close[i] = globalClose[i];
    spread[i] = globalSpread[i];
  }
  __syncthreads(); // wait until all threads have done their bit of Smem loading.

  // now, after doing their part of the data loading, and crucially after the syncthreads, unneeded threads quit, and needed threads calc alpha.
  if (globalTid>=nParameterValues) return;
  float longAlpha = 2.0/(longN[globalTid]+1);
  float shortAlpha = 2.0/(shortN[globalTid]+1);
  
  // initial segment of the sequence. no position is taken - we just compute the emas.
  float longEma = close[0];
  float shortEma = close[0];
  M_repI2(1, takePositionAfterNRecords){
    float priceNow = close[i];
    longEma += longAlpha * (priceNow - longEma);
    shortEma += shortAlpha * (priceNow - shortEma);
  }

  // set up initial position and cash counting.
  // I use floats because those are faster than ints, on GPUs (at least when multiplying with floats)
  float position = (shortEma>=longEma) ? 1 : -1;
  float cashSoFar = - position*close[takePositionAfterNRecords-1] - spread[takePositionAfterNRecords-1]/2;
  
  // the main computation loop
  M_repI2(takePositionAfterNRecords, nRecords-1){
    float priceNow = close[i];
    longEma += longAlpha * (priceNow - longEma);
    shortEma += shortAlpha * (priceNow - shortEma);
    if ((shortEma-longEma) * position <0){ // then change the position
      position = -position;
      cashSoFar -= position*2 * priceNow + spread[i];
    }
  }

  // close final position
  cashSoFar += position * close[nRecords-1] - spread[nRecords-1]/2;

  // write result
  out[globalTid] = cashSoFar;
}


/*
  Parameters:
   float* d_closeValues. device array of close values.
   float* d_spreadValues. device array of spread values.
   int nRecords. The length of the above arrays.
   int takePositionAfterNRecords. The index after which we're going to take the first position.
   int minN & int maxN. The lowest and highest N values that are to be considered.
   int& chosenN and float& profit. The result: the chosen N and the profit for that N.

  Task:
   Try all integer longN and shortN pairs, such that both obey minN and maxN, and such that shortN<longN. Determines which is best.
   Results are written to <chosenLongN>, <chosenShortN>, and <profit>.
*/   
void gpuDoubleEmaMain(float* d_closeValues, float* d_spreadValues, int nRecords, int takePositionAfterNRecords, int minN, int maxN, int& chosenLongN, int& chosenShortN, float& profit){  
  // prep two parameter value arrays in device memory
  int nNvalues = maxN-minN+1; 
  int nParameterValues = (nNvalues*nNvalues - nNvalues)/2;
  int* h_longN = (int*) malloc(nParameterValues*sizeof(int));
  int* h_shortN = (int*) malloc(nParameterValues*sizeof(int));
  int i=0;
  for (int longN=minN; longN<=maxN; longN++) for (int shortN=minN; shortN<longN; shortN++){
    h_longN[i] = longN;
    h_shortN[i] = shortN;
    i++;
  }
  M_cudaMalloc2(d_longN, int, nParameterValues);
  M_cudaMalloc2(d_shortN, int, nParameterValues);
  M_cudaCall(cudaMemcpy(d_longN, h_longN, nParameterValues*sizeof(int), cudaMemcpyHostToDevice));
  M_cudaCall(cudaMemcpy(d_shortN, h_shortN, nParameterValues*sizeof(int), cudaMemcpyHostToDevice));
    
  // kernel execution
  int blockSize = 512; // i don't know what's best, but 512 is my first guess.
  int nBlocks = (nParameterValues+blockSize-1) / blockSize;
  int sharedMemSize = nRecords * 2 * sizeof(float);
  M_cudaMalloc2(d_profit, float, nParameterValues);
  M_cudaCall((gpuDoubleEmaKernel<<<nBlocks, blockSize, sharedMemSize>>>(d_closeValues, d_spreadValues, nRecords, takePositionAfterNRecords, d_longN, d_shortN, nParameterValues, d_profit)));

  // argmax on host (i'll later replace this with argmax on device).
  float* h_profit = (float*) malloc(nParameterValues * sizeof(float));
  M_cudaCall(cudaMemcpy(h_profit, d_profit, nParameterValues*sizeof(float), cudaMemcpyDeviceToHost));
  int chosenParametersIndex = cpuArgmax(h_profit, nParameterValues);
  chosenLongN = h_longN[chosenParametersIndex];
  chosenShortN = h_shortN[chosenParametersIndex];
  profit = h_profit[chosenParametersIndex];

  // debugging output
  //M_repI(nParameterValues) printf("N=%d --> profit = %f.\n", h_parameterValues[i], h_profit[i]);
  
  // finalization
  free(h_longN);
  cudaFree(d_longN);
  free(h_shortN);
  cudaFree(d_shortN);
  free(h_profit);
  cudaFree(d_profit);
}


