#include "cuda_fx.h"

// This file implements optimized GPU MACD.

#define globalEmaElement(emaN, recordIndex) globalEma[(emaN-minN) * nRecords + recordIndex]

__global__ void gpuMacdKernelCalcEma(float* globalClose, float* globalEma, int nRecords, int minN, int maxN){
  int globalTid = threadIdx.x + blockIdx.x * blockDim.x;
  if (globalTid>=maxN-minN+1) return;

  int emaN = globalTid+minN;
  float alpha = 2.0/(emaN+1);

  float ema = globalClose[0];
  globalEmaElement(emaN, 0) = ema;
  M_repI2(1, nRecords){
    ema += alpha * (globalClose[i] - ema);
    globalEmaElement(emaN, i) = ema;
  }
}
  


/*
The optimized MACD main kernel.
 this computes the profit/loss for a particular nLong/nShort setting, optimizing over nSignal, and outputs that optimal profit/loss into a Gmem array.
 The optimization over nSignal is done within the block.
 all threads cooperatively load macd
 then all threads do the computation for one or two nSignal values.
 afterwards, argmax is computed by thread #0, which then writes the result for the block.

parameters:
 float* globalClose is the global memory array of close values.
 float* globalSpread is the global memory array of spread values.
 int nRecords is the number of close values (or, equivalently, the number of spread values).
 int* longN, and int* shortN, are Gmem arrays of parameter values, from which each block handles one.
 float* outProfit and outSignalN are global memory arrays, of length nParameterValues,
   where the final profit/loss & signalN for every parameter setting is to be recorded.

Shared data:
 Arrays:
  s_price (nRecords). this is the close price plus/minus the spread.
  s_macd (nRecords).
  s_profit (nRecords).

There's no space in sMem for spread values.
 I combine the spread values and close values into trading prices.
 At times when macd goes up, the only possible trade is a buy, so the trading price is close+spread/2.
 At times when macd goes down, the only possible trade is a sell, so the trading price is close-spread/2.

*/
__global__ void gpuMacdKernelMain(float* globalClose, float* globalSpread, float* globalEma,
				  int nRecords, int takePositionAfterNRecords,
				  int* longN, int* shortN,
				  int minN, int maxN,
				  float* outProfit, int* outSignalN){
  
  // declaring shared memory. I use "extern __shared__", and split it up in the right pieces.
  extern __shared__ float sharedMemData[];
  float* s_price =   sharedMemData;
  float* s_macd =    s_price + nRecords;
  float* s_profit =  s_macd + nRecords;

  __shared__ float s_lastSpread;
  __shared__ float s_lastClose;
  __shared__ int s_warp0argmaxes[32];

  // loading macd
  for (int i=threadIdx.x; i<nRecords; i+=blockDim.x)
    s_macd[i] = globalEmaElement(shortN[blockIdx.x], i) - globalEmaElement(longN[blockIdx.x], i);
  __syncthreads();
  
  // loading price
  for (int i=threadIdx.x; i<nRecords-1; i+=blockDim.x) if (i>0) s_price[i] = globalClose[i] + globalSpread[i] * ((s_macd[i]>s_macd[i-1]) ? .5 : -.5);
  __syncthreads();

  // loading s_lastSpread & s_lastClose.
  if (threadIdx.x==0) {s_lastSpread = globalSpread[nRecords-1]; s_lastClose = globalClose[nRecords-1];}
  __syncthreads();

  // each thread does zero or more full computations
  for (int parameterValueIndex = threadIdx.x; parameterValueIndex < maxN-minN+1; parameterValueIndex+=blockDim.x){
    
    // prep variables
    float alphaSignal = 2.0/(parameterValueIndex + minN + 1);
    float signal = 0;
    
    // calculate signal without taking position for a while
    M_repI2(1, takePositionAfterNRecords)
      signal += alphaSignal * (s_macd[i] - signal);

    // setup trade variables
    float position = (s_macd[takePositionAfterNRecords-1] > signal) ? 1 : -1;
    float cashSoFar = -position * s_price[takePositionAfterNRecords-1];

    // analysis with buying and selling
    M_repI2(takePositionAfterNRecords, nRecords-1){
      float diff = s_macd[i] - signal; // used twice
      signal += alphaSignal * diff;
      if (diff * position <0){
	// change position
	position = -position;
	cashSoFar -= position*2 * s_price[i];
      }
    
    }

    // close final position & write result
    cashSoFar += position * s_lastClose - s_lastSpread/2;
    s_profit[parameterValueIndex] = cashSoFar;

  }
  __syncthreads();

  // partial argmax by the first warp
  // each thread in the first warp calculates argmax over a subset. thread #0 processes the resulting 32 indices into 1.
  if (threadIdx.x<32){
    int bestIndex = threadIdx.x;
    for (int i=bestIndex+32; i<maxN-minN+1; i+=32)
      if (s_profit[i]>s_profit[bestIndex])
	bestIndex = i;
    s_warp0argmaxes[threadIdx.x] = bestIndex;
  }
  
  // thread #0 calculates the max over the 32 partial maxes, and writes the result.
  if (threadIdx.x==0){
    int bestIndex = s_warp0argmaxes[0];
    M_repI2(1, 32)
      if (s_profit[s_warp0argmaxes[i]] > s_profit[bestIndex])
	bestIndex = s_warp0argmaxes[i];
    outProfit[blockIdx.x] = s_profit[bestIndex];
    outSignalN[blockIdx.x] = bestIndex+minN;
  }
}






/*
  Optimized gpu macd.
  Parameters:
   float* d_closeValues. device array of close values.
   float* d_spreadValues. device array of spread values.
   int nRecords. The length of the above arrays.
   int takePositionAfterNRecords. The index after which we're going to take the first position.
   int minN & int maxN. The lowest and highest N values that are to be considered.
   int& chosenLongN, chosenShortN, chosenSignalN, and float& profit. The result: the chosen N values and the profit for those values.

  Task:
   Try all integer longN, shortN, and signalN triples that obey minN and maxN. Determines which is best.
   Results are written to <chosenLongN>, <chosenShortN>, <chosenSignalN>, and <profit>.
*/   


void gpuMacdMain(float* d_closeValues, float* d_spreadValues,
		 int nRecords, int takePositionAfterNRecords,
		 int minN, int maxN,
		 int& chosenLongN, int& chosenShortN, int& chosenSignalN, float& profit){  
  
  // prep two parameter value arrays in device memory
  int nNvalues = maxN-minN+1; 
  int nParameterValues = (nNvalues*nNvalues - nNvalues)/2;
  int* h_longN = (int*) malloc(nParameterValues*sizeof(int));
  int* h_shortN = (int*) malloc(nParameterValues*sizeof(int));
  int i=0;
 for (int longN=minN; longN<=maxN; longN++) for (int shortN=minN; shortN<longN; shortN++){
    h_longN[i] = longN;
    h_shortN[i] = shortN;
    i++;
  }
  M_cudaMalloc2(d_longN, int, nParameterValues);
  M_cudaMalloc2(d_shortN, int, nParameterValues);
  M_cudaCall(cudaMemcpy(d_longN, h_longN, nParameterValues*sizeof(int), cudaMemcpyHostToDevice));
  M_cudaCall(cudaMemcpy(d_shortN, h_shortN, nParameterValues*sizeof(int), cudaMemcpyHostToDevice));
    
  // prep the ema array in device memory
  M_cudaMalloc2(d_ema, float, (maxN-minN+1)*nRecords);
  int blockSize = 32;
  int nBlocks = (maxN-minN+1)/blockSize+1;
  M_cudaCall((gpuMacdKernelCalcEma<<<nBlocks, blockSize>>>(d_closeValues, d_ema, nRecords, minN, maxN)));
  
  // kernel execution. unfortunately, only 2^16 blocks fit in a 1d grid, so I need to use multiple grids
  int sharedMemSize = nRecords*3 * sizeof(float);
  M_cudaMalloc2(d_profit, float, nParameterValues);
  M_cudaMalloc2(d_signalN, int, nParameterValues);
  int blocksPerGrid = 1<<15;
  int nGrids = (nParameterValues+blocksPerGrid+1) / blocksPerGrid;
  for (int gridNr=0; gridNr<nGrids; gridNr++){
    int offset = gridNr * blocksPerGrid;
    int blocks_thisGrid = min(nParameterValues-offset, blocksPerGrid);
    M_cudaCall((gpuMacdKernelMain<<<blocks_thisGrid, 512, sharedMemSize>>>
		(d_closeValues, d_spreadValues, d_ema, nRecords, takePositionAfterNRecords,
		 d_longN+offset, d_shortN+offset, minN, maxN, d_profit+offset, d_signalN+offset)));
  }	       

  // argmax on host
  float* h_profit = (float*) malloc(nParameterValues * sizeof(float));
  M_cudaCall(cudaMemcpy(h_profit, d_profit, nParameterValues*sizeof(float), cudaMemcpyDeviceToHost));
  int chosenParametersIndex = cpuArgmax(h_profit, nParameterValues);
  chosenLongN = h_longN[chosenParametersIndex];
  chosenShortN = h_shortN[chosenParametersIndex];
  profit = h_profit[chosenParametersIndex];
  M_cudaCall(cudaMemcpy(&chosenSignalN, d_signalN+chosenParametersIndex, sizeof(int), cudaMemcpyDeviceToHost));  

  // debugging output
  //M_repI(nParameterValues) printf("N=%d --> profit = %f.\n", h_parameterValues[i], h_profit[i]);
  
  // finalization
  free(h_longN); cudaFree(d_longN);
  free(h_shortN); cudaFree(d_shortN);
  free(h_profit); cudaFree(d_profit);
  cudaFree(d_signalN);
  cudaFree(d_ema);
}



/*
  time taken:

  loading macd: .18
  loading price: .06
  argmax by warp #0: .1
  argmax by thread #0: .11
  main computation: 18
*/
